/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file id_table.cpp
 *
 * Implements the cid_table class, which stores the message ids from recently
 * received messages (up to klifetime_secs seconds). 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

// #define ID_TABLE_MAIN

#include "id_table.h"
#include "settings.h"
#include "debug.h"


/***************************************************************************
 * INITIALIZE STATIC MEMBERS
 **************************************************************************/
const uint cid_table::kmid_lifetime_secs =
  ((csettings::instance().get<string>("mid_lifetime_secs") == "") ?
   0 : csettings::instance().get<int>("mid_lifetime_secs"));

/***************************************************************************
 * PUBLIC METHODS
 **************************************************************************/
cid_table::cid_table()
{
  cdebug::instance() << acquire("id_table_debug") << "ctor()" << release();
}

cid_table::~cid_table()
{
  cdebug::instance() << acquire("id_table_debug") << "dtor()" << release();
}

void cid_table::add(double id)
{
  cdebug::instance() << acquire("id_table_debug") 
                     << "add(): adding id = " << id
                     << release();

  clock_t timestamp = clock();

  id_map[id] = timestamp;
  id_q.push(make_pair(id, timestamp));
}

bool cid_table::contains(double id) const
{
  cdebug::instance() << acquire("id_table_debug") 
                     << "contains(): looking up id = " << id
                     << release();

  map<double, clock_t>::const_iterator map_it = id_map.find(id);
  return map_it != id_map.end();
}

void cid_table::purge()
{
  cdebug::instance() << acquire("id_table_purge_debug") 
                     << "purge()"
                     << release();

  if (kmid_lifetime_secs == 0)    // lifetime of 0 means don't purge
    return;

  assert(kmid_lifetime_secs > 1);

  clock_t cutoff = clock() - kmid_lifetime_secs * CLOCKS_PER_SEC;
  
  while (!id_q.empty()) {
    pair<double, clock_t> &entry = id_q.front();

    if (entry.second >= cutoff)
      break;

    // TODO: use map::erase that takes a key?
    map<double, clock_t>::iterator it = id_map.find(entry.first);
    assert(it != id_map.end());
    id_map.erase(it);

    id_q.pop();
  }

}


/***************************************************************************
 * TEST MAIN
 **************************************************************************/
#ifdef ID_TABLE_MAIN
int main(int argc, char **argv)
{
/* p4_header hdr;

  hdr.message_id = 4.57;
  hdr.flags = kp4_hello;
  hdr.data_len = 0;

  cmsg *msg = cmsg_factory::instance().new_msg(hdr);
*/
  cid_table::instance().add(4.57);

  assert(cid_table::instance().contains(4.57));

//  hdr.message_id = 5.23;
//  msg =  cmsg_factory::instance().new_msg(hdr);
  assert(!cid_table::instance().contains(5.23));
}
#endif
  
