/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file msg.h
 *
 * Defines the cmsg class, which is an abstract base class for the specific
 * message class. The cmsg class provides methods to send or broadcast a
 * message, process an incoming message, print a message (with operator <<),
 * etc.
 *
 * The cmsg class is abstract so it cannot be instantiated. To create a new
 * message, instantiate one of the classes derived from cmsg (cmsg_*). To create
 * a message from a header (i.e. an incoming message), use cmsg_factory.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef _MSG_H
#define _MSG_H

#include "protocol.h"

#ifdef _WIN32
  #include <winsock.h>
#else
  #include <netinet/in.h>    // for ntohs, etc.
#endif

#ifdef __SVR4
#include <sys/byteorder.h>  // for ntohs, etc. on solaris
#endif

#include <iostream>
#include <string>
using namespace std;


/*****************************************************************************
 * FORWARD DECLARATIONS
 *****************************************************************************/
class cconn;         // so that we can declare it as a friend
class cmsg_factory;  // same


/*****************************************************************************
 * CMSG CLASS
 *****************************************************************************/

class cmsg {
  public:

  /** ctor
   *
   * Takes the destination plugin id and the message data and length. If the
   * message has no data, pass 0 for data and length. All other fields
   * (including flags!) are set to 0.
   *
   * NOTE: The message does *not* make a copy of the data! Be very careful
   * with this. (We can do this because the plugin api send and broadcast
   * calls are blocking.)
   */
  cmsg(ushort plugin_id, const char *data, uint len);

  /** dtor
   *
   * Virtual because instances of derived classes will be deleted from a
   * cmsg *.
   */
  virtual ~cmsg();

  /** operator <<
   *
   * The stream insertion operator for cmsg - prints out the message.
   *
   * NOTE: this is a *non*-member function so that the ostream can be the lhs
   * of the operator<<.
   */
  friend ostream &operator << (ostream &ostr, const cmsg &msg);


  protected:
  friend class cconn;      // for calling recv_socket, process, etc.
  friend class creceiveq;
  friend class cmsg_factory;

  /** !!!!!
   * NOTE: Any class deriving from cmsg MUST redeclare *either* send *or*
   * broadcast to be public!
   * !!!!!*/

  /** send
   *
   * Sends the message to the given user. Returns the cconn that the message
   * will be sent on. The caller can wait on the cconn to block until it is
   * ready to send. If lock is true, the conn_mgr lock is acquired; otherwise,
   * it's not. (This is necessary so that messages (such as DISCOVER) can send
   * messages (such as CONNECT) within their virt_process methods.)
   *
   * If there is no entry for the given user in the LKA, a dummy cconn is
   * created and its status is set to efailed.
   *
   * An assert is raised if the broadcast flag is set.
   *
   * Non-const because it changes the byte ordering.
   */
  virtual const cconn &send(const string &username, bool lock = true);

  /** broadcast
   *
   * Sends the message on all non-transient connections. If lock is true, the
   * conn_mgr lock is acquired; otherwise, it's not. (This is necessary so
   * that broadcast messages can forward themselves from within process.)
   *
   * Non-const because it changes the byte ordering.
   */
  virtual uint broadcast(bool lock = true);

  /** process
   *
   * Handles the message as if it were incoming, taking appropriate action
   * based on the message's type. This may mean sending new messages if the
   * message is a system message, or putting the message on a receive queue if
   * it is a data message.
   *
   * process is protected since only cconn should call it.
   *
   * process returns true if the message was fully processed and should be
   * deleted, false if it was placed on a receive queue (and therefore SHOULD
   * NOT be deleted).
   *
   * NOTE: This method has VIRTUAL BEHAVIOR!!! This method isn't virtual,
   * since it does basic processing that is necessary for all messages. But it
   * calls a pure virtual function to do the message-specific processing, so
   * it should be treated (more or less) like a virtual method.
   *
   * Non-const because it could forward the message.
   */
  bool process();

  /** virt_process
   *
   * This method does the message-specific processing for an incoming message.
   * (except message id handling - the process method does that and then calls
   * virt_process). It's pure virtual, so derived classes must implement it.
   */
  virtual bool virt_process() const = 0;

  /** recv_socket
   *
   * Read data from the given socket into the message's data buffer. The data
   * buffer is allocated once and only once - its size can't be changed.
   * recv_socket can be called multiple times - it will start writing at the
   * point where the last call to recv_socket left off.
   *
   * recv_socket writes the number of data bytes left to read into
   * data_bytes_left after it has read from the socket. This means that when
   * the message has read all of its data, data_bytes_left gets 0. This
   * indicates that the message should be processed (by calling the process
   * method.)
   *
   * NOTE: this should only be called by an instance of cconn!
   */
  uint recv_socket(int sock, uint &data_bytes_left);

  /** ntoh, hton
   *
   * Convert from network to host byte order, and vice versa.
   *
   * NOTE: These are wrappers for convert_byteorder, which is virtual because
   * some system messages have data that also needs to be converted for byte
   * order.
   */
  void ntoh();
  void hton();

  /** convert_byteorder
   *
   * This does the real work of byte-order conversion for messages. (Note that
   * it's virtual because some message types, like msg_hello and msg_discover
   * and msg_lka_ack, override it since they have data of known types.
   */
#if (defined(_WIN32) || defined(__SVR4))
  typedef ushort(*short_fn)(ushort val);
  typedef ulong (*long_fn)(ulong val);

  static ushort my_htons(ushort val) { return htons(val); }
  static ushort my_ntohs(ushort val) { return ntohs(val); }
  static ulong  my_htonl(ulong val)  { return htonl(val); }
  static ulong  my_ntohl(ulong val)  { return ntohl(val); }
#else
  typedef uint16_t  (*short_fn)(short unsigned int val);
  typedef uint32_t  (*long_fn)(unsigned int val);
#endif  // _WIN32 || __SVR4

  virtual void convert_byteorder(short_fn sfn, long_fn lfn);


  /** ctor
   *
   * Protected so that only the message factory can create a message object
   * from a P4 header.
   */
  cmsg(const p4_header &hdr);

  /** copy ctor
   *
   * Protected so that receiveq can pull them off of a vector.
   */

  cmsg(const cmsg &msg);


  private:
  // prohibit assignment
  cmsg &operator= (const cmsg &msg);

  /**************************************************************************
   * MEMBERS
   **************************************************************************/
  protected:
  p4_header  hdr;
  ulong host_data_len;  // data length, *always* in host byte order!
  char *data;           // 0 if unallocated (so ctors should initialize to 0!)
  uint data_bytes_left;
  bool incoming;

  typedef enum ebyte_order_tt { enetwork, ehost } ebyte_order_t;
  ebyte_order_t  byte_order;
};


#endif  // _MSG_H





